@@ -1,4 +1,4 @@
-// $Id: main.c,v 1.5 2002/06/12 09:47:57 ensc Exp $    --*- c++ -*--
+// $Id: main.c,v 1.12 2002/07/11 17:02:20 ensc Exp $    --*- c++ -*--
 
 // Copyright (C) 2002 Enrico Scholz <enrico.scholz@informatik.tu-chemnitz.de>
 //  
@@ -20,57 +20,90 @@
 #  include <config.h>
 #endif
 
+#include "splint_compat.h"
+
 #include <stdlib.h>
 #include <fcntl.h>
 #include <unistd.h>
 #include <string.h>
-#include <sys/types.h>
 #include <sys/time.h>
-#include <sys/ioctl.h>
+#include <sys/types.h>
 #include <net/if.h>
 #include <sys/socket.h>
+#include <sys/uio.h>
 #include <assert.h>
 #include <sys/select.h>
 #include <sys/param.h>
 #include <errno.h>
 #include <netpacket/packet.h>
 #include <net/if_arp.h>
+#include <sys/ioctl.h>
 
-
-#include "config.h"
+#include "cfg.h"
+#include "wrappers.h"
 #include "dhcp.h"
 #include "inet.h"
-#include "wrappers.h"
 #include "recvfromflags.h"
 
 #include "assertions.h"
 #include "logging.h"
 
-static struct ServerInfoList	servers;
-static struct FdInfoList	fds;
+  //#define ENABLE_AGENT_REPLACE		1
+
+typedef enum {
+  acIGNORE,		//< Do nothing...
+  acREMOVE_ID,		//< Remove an already existing agent-id field
+  acADD_ID		//< Add an agent-id field if such a field does not
+			//< already exists
+#ifdef ENABLE_AGENT_REPLACE  
+  ,acREPLACE_ID		//< Replace an already existing agent-id field
+#endif  
+} OptionFillAction;
+
+/*@checkmod@*/static struct ServerInfoList	servers;
+/*@checkmod@*/static struct FdInfoList	fds;
 
 inline static void
-fillFDSet(/*@out@*/fd_set			*fd_set,
+fillFDSet(/*@out@*/fd_set			*fdset,
 	  /*@out@*/int				*max)
+    /*@globals fds@*/
+    /*@modifies *fdset, *max@*/
 {
-  size_t	i = 0;
-  FD_ZERO(fd_set);
+    /*@-nullptrarith@*/
+  struct FdInfo const *		fdinfo;
+  struct FdInfo const * const	end_fdinfo = fds.dta+fds.len;
+    /*@=nullptrarith@*/
+
+  assert(fds.dta!=0 || fds.len==0);
+
+  FD_ZERO(fdset);
   *max = -1;
-  
-  for (i=0; i<fds.len; ++i) {
-    *max = MAX(*max, fds.dta[i].fd);
-    FD_SET(fds.dta[i].fd, fd_set);
+
+  for (fdinfo=fds.dta; fdinfo<end_fdinfo; ++fdinfo) {
+    assert(fdinfo!=0);
+    
+    *max = MAX(*max, fdinfo->fd);
+    FD_SET(fdinfo->fd, fdset);
   }
 }
 
-inline static struct FdInfo const *
+inline static /*@exposed@*//*@null@*/struct FdInfo const *
 lookupFD(/*@in@*/struct in_addr const addr)
+    /*@globals fds@*/
 {
-  size_t		i;
+    /*@-nullptrarith@*/
+  struct FdInfo const *			fdinfo;
+  struct FdInfo const * const		end_fdinfo = fds.dta+fds.len;
+    /*@=nullptrarith@*/
+
+  assert(fds.dta!=0 || fds.len==0);
   
-  for (i=0; i<fds.len; ++i) {
-    struct FdInfo const	* const		fd = &fds.dta[i];
-    if (fd->iface->if_ip==addr.s_addr) return fd;
+  for (fdinfo=fds.dta; fdinfo<end_fdinfo; ++fdinfo) {
+    assert(fdinfo!=0);
+    
+      /* We must check for both the real and "faked" giaddr IP here */
+    if (fdinfo->iface->if_real_ip==addr.s_addr ||
+	fdinfo->iface->if_ip     ==addr.s_addr) return fdinfo;
   }
 
   return 0;
@@ -78,37 +111,53 @@ lookupFD(/*@in@*/struct in_addr const addr)
 
 inline static size_t
 determineMaxMTU()
+    /*@globals fds@*/
+    /*@modifies@*/
 {
-  size_t		result = 1500;
-  struct FdInfo	const	*fd;
+  size_t				result = 1500;
+    /*@-nullptrarith@*/
+  struct FdInfo const *			fdinfo;
+  struct FdInfo const * const		end_fdinfo = fds.dta+fds.len;
+    /*@=nullptrarith@*/
+
+  assert(fds.dta!=0 || fds.len==0);
   
-  for (fd=fds.dta; fd<fds.dta+fds.len; ++fd)
-    result = MAX(result, fd->iface->if_mtu);
+  for (fdinfo=fds.dta; fdinfo<end_fdinfo; ++fdinfo) {
+    assert(fdinfo!=0);
+    
+    result = MAX(result, fdinfo->iface->if_mtu);
+  }
 
   return result;
 }
 
 inline static bool
-isValidHeader(struct DHCPHeader *header)
+isValidHeader(/*@in@*/struct DHCPHeader *header)
+    /*@globals internalState@*/
+    /*@modifies internalState@*/
 {
   char const		*reason = 0;
-  unsigned int		mbz = header->flags.mbz1+header->flags.mbz2;
- 
-  if      (mbz!=0)                 { reason = "Invalid flags field\n"; }
-  else if (header->hops>=MAX_HOPS) { reason = "Looping detected\n"; }
+
+    /*@-strictops@*/
+  if ((header->flags&~flgDHCP_BCAST)!=0) { reason = "Invalid flags field"; }
+    /*@=strictops@*/
+  else if (header->hops>=MAX_HOPS)       { reason = "Looping detected"; }
 #if 0  
   else switch (header->htype) {
     case ARPHRD_ETHER	:
-      if (header->hlen!=ETH_ALEN) {  reason = "Invalid hlen for ethernet\n"; }
+      if (header->hlen!=ETH_ALEN) {  reason = "Invalid hlen for ethernet"; }
       break;
     default		:
       break;	// Not active handled by us; will be forwarded as-is
   }
-#endif  
+#else
+  else {}
+#endif
+  
   if (reason==0) switch (header->op) {
     case opBOOTREPLY	:
     case opBOOTREQUEST	:  break;
-    default		:  reason = "Unknown operation\n"; break;
+    default		:  reason = "Unknown operation"; break;
   };
 
   if (reason!=0) LOGSTR(reason);
@@ -119,15 +168,16 @@ isValidHeader(struct DHCPHeader *header)
 inline static bool
 isValidOptions(/*@in@*/struct DHCPOptions const	*options,
 	       size_t				o_len)
+    /*@*/
 {
   bool				seen_end     = false;
   struct DHCPSingleOption const	*opt         = reinterpret_cast(struct DHCPSingleOption const *)(options->data);
-  struct DHCPSingleOption const	*end_options = reinterpret_cast(struct DHCPSingleOption const *)(reinterpret_cast(uint8_t const *)(options) + o_len);
+  struct DHCPSingleOption const	*end_options = reinterpret_cast(struct DHCPSingleOption const *)(&reinterpret_cast(uint8_t const *)(options)[o_len]);
 
     /* Is this really ok? Is an empty option-field RFC compliant? */
   if (o_len==0) return true;
   if (o_len<=4) return false;
-  if (options->cookie != DHCP_COOKIE) return false;
+  if (options->cookie != optDHCP_COOKIE) return false;
 
   do {
     switch (opt->code) {
@@ -137,34 +187,17 @@ isValidOptions(/*@in@*/struct DHCPOptions const	*options,
 
     opt = DHCP_nextSingleOptionConst(opt);
   } while (opt < end_options);
-
+  
   return (seen_end && opt==end_options);
 }
 
 inline static size_t
-fillOptions(void				*option_ptr,
-	    /*@in@*/struct InterfaceInfo const	*iface)
+addAgentOption(/*@in@*/struct InterfaceInfo const * const	iface,
+	       struct DHCPSingleOption				*end_opt,
+	       size_t						len)
+    /*@modifies *end_opt@*/
 {
-  struct DHCPSingleOption 		*opt     = static_cast(struct DHCPSingleOption *)(option_ptr);
-  struct DHCPSingleOption		*end_opt = 0, *relay_opt = 0;
-  size_t				len;
-  
-  do {
-    switch (opt->code) {
-      case cdRELAY_AGENT	:  relay_opt = opt; break;
-      case cdEND		:  end_opt   = opt; break;
-      default			:  break;
-    }
-
-    opt = DHCP_nextSingleOption(opt);
-  } while (end_opt==0);
-
-    /* Determine used space until end-tag and add space for the end-tag itself
-     * (1 octet). */
-  len  = (reinterpret_cast(char *)(end_opt) -
-	  static_cast(char *)(option_ptr) + 1u);
-  
-    /* Check if a relay-agent field exists already; if not, replace the end-tag
+    /* Replace the end-tag
        
      *  - - - - - - -----
      * |           | END |
@@ -175,49 +208,133 @@ fillOptions(void				*option_ptr,
      * |           |     |     | opt | len |            |     |
      *  - - - - - - ----- ----- ----- ----- ------------ -----
      * */
-  if (relay_opt==0) {
-    assert(strlen(iface->aid)<=IFNAMSIZ);
+  
+  assert(strlen(iface->aid)<=IFNAMSIZ);
     
-      /* Add space needed for our RFC 3046 agent id. See figure above for
-       * details. */
-    len += 4 + strlen(iface->aid);
+    /* Add space needed for our RFC 3046 agent id. See figure above for
+     * details. */
+  len += 4 + strlen(iface->aid);
 
-      /* 'len' should now have the length of the complete option-field. RFC 2131
-       * sets a lower limit of 312 octets, so we are checking against this
-       * value. Since the function got only the real options without the
-       * magic-cookie, 4 octets must be added.
-       *
-       * Because the underlying buffer was declared to hold more than this
-       * minimum amount, we can exclude overflows here.
-       *
-       * Further versions of this software should make it possible to configure
-       * the maximum size at runtime. */
-    if (len+4 < 312) {
-	/* replace old end-tag with our information */
-      end_opt->code    = cdRELAY_AGENT;
-      end_opt->data[0] = agCIRCUITID;	/* code for circuit id as specified by RFC 3046 */
-      end_opt->data[1] = static_cast(uint8_t)(strlen(iface->aid));
-      end_opt->len     = 2 + end_opt->data[1];
-      memcpy(end_opt->data+2, iface->aid, end_opt->data[1]);
+    /* 'len' should now have the length of the complete option-field. RFC 2131
+     * sets a lower limit of 312 octets, so we are checking against this
+     * value. Since the function got only the real options without the
+     * magic-cookie, 4 octets must be added.
+     *
+     * Because the underlying buffer was declared to hold more than this
+     * minimum amount, we can exclude overflows here.
+     *
+     * Further versions of this software should make it possible to configure
+     * the maximum size at runtime. */
+  if (len+4 < 312) {
+      /* replace old end-tag with our information */
+    end_opt->code    = cdRELAY_AGENT;
+    end_opt->data[0] = agCIRCUITID;	/* circuit id code as specified by RFC 3046 */
+    end_opt->data[1] = static_cast(uint8_t)(strlen(iface->aid));
+    end_opt->len     = 2 + end_opt->data[1];
+    memcpy(&end_opt->data[2], iface->aid, end_opt->data[1]);
 
-	/* set new end-tag */
-      end_opt = DHCP_nextSingleOption(end_opt);
-      end_opt->code = cdEND;
-    }
+      /* set new end-tag */
+    end_opt = DHCP_nextSingleOption(end_opt);
+    end_opt->code = cdEND;
   }
 
   return len;
 }
 
+#ifdef ENABLE_AGENT_REPLACE
+inline static size_t
+replaceAgentOption(/*@in@*/struct InterfaceInfo const * const	iface,
+		   struct DHCPSingleOption			*relay_opt,
+		   struct DHCPSingleOption			*end_opt,
+		   size_t					len)
+{
+  size_t	opt_len = DHCP_getOptionLength(relay_opt);
+  size_t	str_len = strlen(iface->aid);
+
+  assert(relay_opt!=0);
+  
+  if (str_len+4<=opt_len) {
+    relay_opt->len     = str_len + 2;
+    relay_opt->data[1] = static_cast(uint8_t)(str_len);
+    memcpy(relay_opt->data+2, iface->aid, str_len);
+
+    if (str_len+4<opt_len)
+      memset(relay_opt->data+2+str_len, /*@+charint@*/cdPAD/*@=charint@*/, opt_len-str_len-4);
+  }
+  else {
+    DHCP_zeroOption(relay_opt);
+    len = addAgentOption(iface, end_opt, len);
+  }
+
+  return len;
+}
+#endif
+
+  /*@-mustmod@*/
+inline static size_t
+fillOptions(/*@in@*/struct InterfaceInfo const* const	iface,
+	    /*@dependent@*/void				*option_ptr,
+	    OptionFillAction				action)
+    /*@modifies *option_ptr@*/
+{
+  /*@dependent@*/struct DHCPSingleOption 	*opt       = static_cast(struct DHCPSingleOption *)(option_ptr);
+  /*@dependent@*/struct DHCPSingleOption	*end_opt   = 0;
+  /*@dependent@*/struct DHCPSingleOption	*relay_opt = 0;
+  size_t			len;
+
+  do {
+    switch (opt->code) {
+      case cdRELAY_AGENT	:
+	if (opt->data[0]==agCIRCUITID) relay_opt = opt;
+	break;
+      case cdEND		:  end_opt   = opt; break;
+      default			:  break;
+    }
+
+    opt = DHCP_nextSingleOption(opt);
+  } while (end_opt==0);
+
+  assert(end_opt>=option_ptr);
+
+    /* Determine used space until end-tag and add space for the end-tag itself
+     * (1 octet). */
+    /*@-strictops@*/
+  len  = (reinterpret_cast(char *)(end_opt) -
+	  static_cast(char *)(option_ptr) + 1u);
+    /*@=strictops@*/
+
+  switch (action) {
+    case acREMOVE_ID	:
+      if (relay_opt!=0) DHCP_zeroOption(relay_opt);
+      break;
+    case acADD_ID	:
+      if (relay_opt==0) len = addAgentOption(iface, end_opt, len);
+      break;
+#ifdef ENABLE_AGENT_REPLACE      
+    case acREPLACE_ID	:
+      if (relay_opt==0) len = addOption(end_opt, len);
+      else              len = replaceAgentOption(iface, relay_opt, end_opt, len);
+      break;
+#endif
+    case acIGNORE	:  break;
+    default		:  assert(false);
+  }
+      
+  return len;
+}
+  /*@=mustmod@*/
+
 inline static uint16_t
 calculateCheckSum(/*@in@*/void const * const	dta,
 		  size_t size,
 		  uint32_t sum)
+    /*@*/
 {
   size_t		i;
   uint16_t const	*data = reinterpret_cast(uint16_t const *)(dta);
   
   for (i=0; i<size/2; ++i) sum += ntohs(data[i]);
+  
   if (size%2 != 0) {
     union {
 	uint8_t		aval[2];
@@ -238,6 +355,7 @@ calculateCheckSum(/*@in@*/void const * const	dta,
 
 inline static void
 fixCheckSumIP(struct iphdr * const	ip)
+    /*@modifies *ip@*/
 {
   ip->check = 0;
   ip->check = htons(~calculateCheckSum(ip, sizeof(*ip), 0));
@@ -246,7 +364,8 @@ fixCheckSumIP(struct iphdr * const	ip)
 static void
 fixCheckSumUDP(struct udphdr * const			udp,
 	       /*@in@*/struct iphdr const * const	ip,
-	       void const * const			data)
+	       /*@in@*/void const * const		data)
+    /*@modifies *udp@*/
 {
   uint32_t		sum;
   struct {
@@ -255,10 +374,13 @@ fixCheckSumUDP(struct udphdr * const			udp,
       uint8_t		mbz;
       uint8_t		proto;
       uint16_t		len;
-  } __attribute__((__packed__))	const pseudo_hdr = {
-    ip->saddr, ip->daddr,
-    0,
-    ip->protocol, udp->len };
+  } __attribute__((__packed__))	pseudo_hdr;
+
+  pseudo_hdr.src   = ip->saddr;
+  pseudo_hdr.dst   = ip->daddr;
+  pseudo_hdr.mbz   = 0;
+  pseudo_hdr.proto = ip->protocol;
+  pseudo_hdr.len   = udp->len;
 
   udp->check = 0;
   sum = calculateCheckSum(&pseudo_hdr, sizeof pseudo_hdr, 0);
@@ -271,38 +393,43 @@ fixCheckSumUDP(struct udphdr * const			udp,
 }
 
 inline static void
-sendEtherFrame(struct InterfaceInfo const	*iface,
-	       struct DHCPllPacket		*frame,
-	       char const			*buffer,
-	       size_t				size)
+sendEtherFrame(/*@in@*/struct InterfaceInfo const	*iface,
+	       /*@dependent@*/struct DHCPllPacket	*frame,
+	       /*@dependent@*//*@in@*/char const	*buffer,
+	       size_t					size)
+    /*@globals internalState, fds@*/
+    /*@modifies internalState, *frame@*/
 {
   struct sockaddr_ll		sock;
   struct msghdr			msg;
-  struct iovec			iovec_data[2];
+  /*@temp@*/struct iovec	iovec_data[2];
+    /*@-sizeoftype@*/
+  size_t const			szUDPHDR = sizeof(struct udphdr);
+    /*@=sizeoftype@*/
 
     /* We support ethernet only and the config-part shall return ethernet-macs
      * only... */
   assert(iface->if_maclen == sizeof(frame->eth.ether_dhost));
   
   memset(&sock, 0, sizeof sock);
-  sock.sll_family    = AF_PACKET;
+  sock.sll_family    = static_cast(sa_family_t)(AF_PACKET);
   sock.sll_ifindex   = static_cast(int)(iface->if_idx);
     /* We do not need to initialize the other attributes of rcpt_sock since
      * dst-hwaddr et.al. are determined by the ethernet-frame defined below */
 
   memcpy(frame->eth.ether_shost, iface->if_mac,  iface->if_maclen);
   
-  frame->ip.version  = 4;
+  frame->ip.version  = 4u;
   frame->ip.ihl      = sizeof(frame->ip)/4u;
   frame->ip.tos      = 0;
-  frame->ip.tot_len  = htons(sizeof(frame->ip) + sizeof(struct udphdr) + size);
+  frame->ip.tot_len  = htons(sizeof(frame->ip) + szUDPHDR + size);
   frame->ip.id       = 0;
   frame->ip.frag_off = htons(IP_DF);
   frame->ip.ttl      = 64;
   frame->ip.protocol = IPPROTO_UDP;
   frame->ip.saddr    = iface->if_ip;
 
-  frame->udp.len     = htons(sizeof(struct udphdr) + size);
+  frame->udp.len     = htons(szUDPHDR + size);
 
   fixCheckSumIP(&frame->ip);
   fixCheckSumUDP(&frame->udp, &frame->ip, buffer);
@@ -320,14 +447,18 @@ sendEtherFrame(struct InterfaceInfo const	*iface,
   msg.msg_controllen = 0;
   msg.msg_flags      = 0;
 
+  assertDefined(msg.msg_iov);
+
   Wsendmsg(fds.raw_fd, &msg, 0);
 }
 
 inline static void
 sendToClient(/*@in@*/struct FdInfo const * const	fd,
 	     /*@in@*/struct DHCPHeader const * const	header,
-	     /*@in@*/char const * const			buffer,
+	     /*@in@*//*@dependent@*/char const * const	buffer,
 	     size_t const				size)
+    /*@globals internalState, fds@*/
+    /*@modifies internalState@*/
 {
   struct DHCPllPacket		frame;
   struct InterfaceInfo const	*iface = fd->iface;
@@ -345,7 +476,7 @@ sendToClient(/*@in@*/struct FdInfo const * const	fd,
   else
     memset(frame.eth.ether_dhost, 255,            sizeof frame.eth.ether_dhost);
   
-  if (!header->flags.bcast && header->ciaddr!=0)
+  if ((header->flags&flgDHCP_BCAST)!=0 && header->ciaddr!=0)
     frame.ip.daddr  = header->ciaddr;
   else if (iface->allow_bcast)
     frame.ip.daddr  = INADDR_BROADCAST;
@@ -359,9 +490,11 @@ sendToClient(/*@in@*/struct FdInfo const * const	fd,
 }
 
 inline static void
-sendServerBcast(struct ServerInfo const	* const		server,
-		char const * const			buffer,
-		size_t const				size)
+sendServerBcast(/*@in@*/struct ServerInfo const	* const		server,
+		/*@dependent@*//*@in@*/char const * const	buffer,
+		size_t const					size)
+    /*@globals internalState, fds@*/
+    /*@modifies internalState@*/
 {
   struct DHCPllPacket		frame;
   struct InterfaceInfo const	*iface = server->info.iface;
@@ -380,9 +513,11 @@ sendServerBcast(struct ServerInfo const	* const		server,
 }
 
 inline static void
-sendServerUnicast(struct ServerInfo const * const	server,
-		  char const * const			buffer,
-		  size_t const				size)
+sendServerUnicast(/*@in@*/struct ServerInfo const * const	server,
+		  /*@in@*/char const * const			buffer,
+		  size_t const					size)
+    /*@globals internalState, fds@*/
+    /*@modifies internalState@*/
 {
   struct sockaddr_in	sock;
 
@@ -398,14 +533,21 @@ sendServerUnicast(struct ServerInfo const * const	server,
 }
 
 inline static void
-sendToServer(char const * const				buffer,
+sendToServer(/*@in@*//*@dependent@*/char const * const	buffer,
 	     size_t const				size)
+    /*@globals servers, fds, internalState@*/
+    /*@modifies internalState@*/
 {
-  size_t		j;
+    /*@-nullptrarith@*/
+  struct ServerInfo const *		server;
+  struct ServerInfo const * const	end_server = servers.dta+servers.len;
+    /*@=nullptrarith@*/
 
-  for (j=0; j<servers.len; ++j) {
-    struct ServerInfo const * const	server = servers.dta + j;
+  assert(servers.len==0 || servers.dta!=0);
 
+  for (server=servers.dta; server<end_server; ++server) {
+    assert(server!=0);
+    
     switch (server->type) {
       case svUNICAST	:
 	sendServerUnicast(server, buffer, size);
@@ -424,9 +566,11 @@ handlePacket(/*@in@*/struct FdInfo const * const		fd,
 	     /*@in@*/struct InterfaceInfo const * const		iface_orig,
 	     /*@dependent@*/char * const			buffer,
 	     size_t						size)
+    /*@globals fds, servers, internalState@*/
+    /*@modifies internalState@*/
 {
   struct DHCPHeader * const	header  = reinterpret_cast(struct DHCPHeader *)(buffer);
-  struct DHCPOptions * const	options = reinterpret_cast(struct DHCPOptions *)(buffer + sizeof(*header));
+  struct DHCPOptions * const	options = reinterpret_cast(struct DHCPOptions *)(&buffer[sizeof(*header)]);
   size_t			options_len = size - sizeof(*header);
   
 
@@ -434,20 +578,29 @@ handlePacket(/*@in@*/struct FdInfo const * const		fd,
   if (!isValidHeader(header)) return;
   ++header->hops;
 
-    /* Check if we are the first agent; if so, set 'giaddr' and the relay-agent
-     * field, else do not touch the packet */
-  if (header->giaddr==0) {
+    /* Check if we are the first agent or if we handle a packet indented for us;
+     * if so, set 'giaddr' and the relay-agent field, else do not touch the
+     * packet */
+  if (header->giaddr==0 || header->giaddr==fd->iface->if_ip) {
     header->giaddr = fd->iface->if_ip;
   
     if (!isValidOptions(options, options_len)) {
-      LOG("Invalid options\n");
+      LOG("Invalid options");
       return;
     }
 
-      /* Do not add an option-field if such a field does not already exists. */
+      /* Do not fill the option-field if this field does not exists. */
     if (options_len!=0) {
+      OptionFillAction		action;
+
+      switch (header->op) {
+	case opBOOTREPLY	:  action = acREMOVE_ID; break;
+	case opBOOTREQUEST	:  action = acADD_ID;    break;
+	default			:  assert(false); action = acIGNORE; break;
+      }
+      
 	/* Fill agent-info and adjust size-information */
-      options_len  = fillOptions(options->data, fd->iface);
+      options_len  = fillOptions(fd->iface, options->data, action);
       options_len += sizeof(options->cookie);
       size         = options_len + sizeof(*header);
     }
@@ -457,12 +610,12 @@ handlePacket(/*@in@*/struct FdInfo const * const		fd,
 
   switch (header->op) {
     case opBOOTREPLY	:
-      if      (!iface_orig->has_servers) LOG("BOOTREPLY request from interface without servers\n");
-      else if (!fd->iface->has_clients)  LOG("BOOTREPLY request for interface without clients\n");
+      if      (!iface_orig->has_servers) LOG("BOOTREPLY request from interface without servers");
+      else if (!fd->iface->has_clients)  LOG("BOOTREPLY request for interface without clients");
       else sendToClient(fd, header, buffer, size);
       break;
     case opBOOTREQUEST	:
-      if (!iface_orig->has_clients)      LOG("BOOTREQUEST request from interface without clients\n");
+      if (!iface_orig->has_clients)      LOG("BOOTREQUEST request from interface without clients");
       else sendToServer(buffer, size);
       break;
       
@@ -475,33 +628,43 @@ handlePacket(/*@in@*/struct FdInfo const * const		fd,
 /*@noreturn@*/
 inline static void
 execRelay()
+    /*@globals fds, servers, internalState@*/
+    /*@modifies internalState@*/
 {
   size_t const			max_mtu   = determineMaxMTU();
-  size_t const			total_len = max_mtu + IFNAMSIZ + 4;
-  char				*buffer   = 0;
+  size_t const			len_total = max_mtu + IFNAMSIZ + 4;
+  char				*buffer   = static_cast(char *)(Ealloca(len_total));
+    /*@-sizeoftype@*/
+  size_t const			szDHCPHDR = sizeof(struct DHCPHeader);
+    /*@=sizeoftype@*/
 
-  buffer = static_cast(char *)(alloca(total_len));
+  assert(buffer!=0);
+  assert(fds.dta!=0 || fds.len==0);
 
   while (true) {
-    fd_set			fd_set;
+    fd_set			fdset;
     int				max;
-    size_t			i;
+      /*@-nullptrarith@*/
+    struct FdInfo const *	fdinfo;
+    struct FdInfo const * const end_fdinfo = fds.dta+fds.len;
+      /*@=nullptrarith@*/
     
-    fillFDSet(&fd_set, &max);
-    if (Wselect(max+1, &fd_set, 0, 0, 0)==-1) continue;
+    fillFDSet(&fdset, &max);
+    if /*@-type@*/(Wselect(max+1, &fdset, 0, 0, 0)==-1)/*@=type@*/ continue;
 
-    for (i=0; i<fds.len; ++i) {
-      struct FdInfo const	*fd_info = &fds.dta[i];
-      size_t			size;
-      struct sockaddr_in	addr;
-      struct in_pktinfo		pkinfo;
-      int			flags = 0;
+    for (fdinfo=fds.dta; fdinfo<end_fdinfo; ++fdinfo) {
+      size_t				size;
+      struct sockaddr_in		addr;
+      struct in_pktinfo			pkinfo;
+      int				flags = 0;
+
+      assert(fdinfo!=0);
       
-      if (!FD_ISSET(fd_info->fd, &fd_set)) continue;
+      if (!FD_ISSET(fdinfo->fd, &fdset)) /*@innercontinue@*/continue;
 
 	/* Is this really correct? Can we receive fragmented IP datagrams being
 	 * assembled larger than the MTU of the attached interfaces? */
-      size = WrecvfromFlagsInet4(fd_info->fd, buffer, total_len, &flags,
+      size = WrecvfromFlagsInet4(fdinfo->fd, buffer, len_total, &flags,
 				 &addr, &pkinfo);
 
 #ifdef ENABLE_LOGGING      
@@ -509,15 +672,17 @@ execRelay()
 #endif
       
       if (static_cast(ssize_t)(size)==-1) {
-	int		error = errno;
-	LOG("recvfrom(): ");
-	LOGSTR(strerror(error));
+	char const *	msg = strerror(errno);
+	  
+	LOG("recvfrom() failed");
+	LOGSTR(msg);
       }
-      else if (size < sizeof(struct DHCPHeader)) {
-	LOG("Malformed package\n");
+      else if (size < szDHCPHDR) {
+	LOG("Malformed package");
       }
       else {
-	struct InterfaceInfo const	*iface_orig = fd_info->iface;
+	struct InterfaceInfo const * const	iface_orig = fdinfo->iface;
+	struct FdInfo const *			fd_real    = fdinfo;
 	
 	  /* Broadcast messages are designated for the interface, so lookup the
 	   * "real" dest-interface for unicast-messages only.
@@ -525,14 +690,14 @@ execRelay()
 	   * \todo: Can we setup the routing that datagrams from the server are
 	   *        received by the giaddr-interface? */
 	if (pkinfo.ipi_addr.s_addr!=INADDR_BROADCAST)
-	  fd_info = lookupFD(pkinfo.ipi_addr);
+	  fd_real = lookupFD(pkinfo.ipi_addr);
 
-	if (fd_info==0) LOG("Received package on unknown interface\n");
-	else if (size > fd_info->iface->if_mtu) {
-	  LOG("Unexpected large packet\n");
+	if (fd_real==0) LOG("Received package on unknown interface");
+	else if (size > fd_real->iface->if_mtu) {
+	  LOG("Unexpected large packet");
 	}
 	else 
-	  handlePacket(fd_info, iface_orig, buffer, size);
+	  handlePacket(fd_real, iface_orig, buffer, size);
       }
     }
   }
@@ -540,16 +705,23 @@ execRelay()
 
 int
 main(int argc, char *argv[])
+    /*@globals fds, servers, internalState, fileSystem@*/
+    /*@modifies fds, servers, internalState, fileSystem@*/
 {
   struct InterfaceInfoList		ifs;
 
   checkCompileTimeAssertions();
 
+    /* We create a mem-leak here, but this is not important because the parent
+     * exits immediately and frees the resources and the child never exits */
+    /*@-compdestroy@*//*@-superuser@*/
   switch (initializeSystem(argc, argv, &ifs, &servers, &fds)) {
+    /*@=superuser@*/
     case -1	:  return 5;
     case 0	:  execRelay();
     default	:  return 0;
   }
+    /*@=compdestroy@*/  
 }
 
   // Local Variables:
